from datetime import datetime, timedelta
from .operation.SqlOperation import DescTableOperation


class TableStructureCache:
    __cache = None
    __cache_timeout = None

    def __init__(self, timeout: timedelta):
        self.__cache = {}
        self.__cache_timeout = timeout

    def cache_structure(self, table: str, columns: (set, list)):
        structure = TableStructure(table, columns)
        self.__cache[table] = structure
        return structure

    def find_structure(self, table: str, cursor):
        structure = self.__cache.get(table)
        if structure is not None and not structure.expired(self.__cache_timeout):
            return structure
        op = DescTableOperation(table)
        structure = op.operate(cursor)
        columns = set([field['field'] for field in structure])
        structure = self.cache_structure(table, columns)
        return structure


class TableStructure:
    table_name = None
    __columns = None
    __init_time = None

    def __init__(self, table_name: str, columns: (set, list)):
        self.table_name = table_name
        self.__columns = set(columns)
        self.__init_time = datetime.now()

    def exists_columns(self, obj: dict):
        return list(self.__columns & set(obj.keys()))

    def expired(self, timeout: timedelta):
        return self.__init_time + timeout < datetime.now()



import logging
from abc import ABCMeta, abstractmethod

from ..sqlbuilder import SqlBuilder
from .SqlOperation import *
from ..TableCache import TableStructureCache

logger = logging.getLogger('ezmysql')

__all__ = ['TableOperation', 'TableInsertOperation', 'UpdateTableOperation', 'SelectTableSqlOperation',
           'SelectObjectFromTableOperation', 'SelectOneFromTableOperation', 'DeleteTableOperation',
           'DescTableOperation']


class TableOperation(SqlOperation):
    __metadata__ = ABCMeta
    _table_cache: TableStructureCache = None

    def __init__(self, cache: TableStructureCache):
        self._table_cache = cache

    @abstractmethod
    def operate(self, cursor):
        pass


class TableInsertOperation(TableOperation):
    __table_name = None
    __item = None

    def __init__(self, cache: TableStructureCache, table_name: str, item: dict):
        super().__init__(cache)
        self.__table_name = table_name
        self.__item = item

    def operate(self, cursor):
        structure = self._table_cache.find_structure(self.__table_name, cursor)

        exists_columns = structure.exists_columns(self.__item)
        sql = SqlBuilder.insert_into(self.__table_name).values(exists_columns).build()
        values = [self.__item[key] for key in exists_columns]
        logger.debug('executing sql [{}] with params \n{}'.format(sql, ','.join([str(param) for param in values])))
        update_operation = UpdateSqlOperation(sql, values)
        update_operation.operate(cursor)
        return cursor.lastrowid


class UpdateTableOperation(TableOperation):
    __table_name = None
    __content = None
    __select_param = None

    def __init__(self, cache: TableStructureCache, table_name: str, content: dict, where: dict):
        super().__init__(cache)
        self.__table_name = table_name
        self.__content = content
        self.__select_param = where

    def operate(self, cursor):
        table_structure = self._table_cache.find_structure(self.__table_name, cursor)
        exists_content_columns = table_structure.exists_columns(self.__content)
        exists_param_columns = table_structure.exists_columns(self.__select_param)
        if len(exists_param_columns) == 0:
            raise Exception('invalid operation: where params is null')
        params = [self.__content[key] for key in exists_content_columns] + \
                 [self.__select_param[key] for key in exists_param_columns if self.__select_param[key] is not None]
        sql = SqlBuilder.update(self.__table_name).set(exists_content_columns) \
            .where_equals(exists_param_columns, self.__select_param).build()
        logger.debug('executing sql [{}] with params \n{}'.format(sql, ','.join([str(param) for param in params])))
        cursor.execute(sql, params)
        return cursor.rowcount


class DeleteTableOperation(TableOperation):
    __table_name: str = None
    __params: dict = None

    def __init__(self, cache: TableStructureCache, table_name: str, where: dict):
        super().__init__(cache)
        self.__table_name = table_name
        self.__params = where

    def operate(self, cursor):
        table_structure = self._table_cache.find_structure(self.__table_name, cursor)
        exists_param_columns = table_structure.exists_columns(self.__params)
        if len(exists_param_columns) == 0:
            raise Exception('invalid operation: where params is null')
        params = [self.__params[key] for key in exists_param_columns if self.__params[key] is not None]
        sql = SqlBuilder.delete_from(self.__table_name).where_equals(exists_param_columns, self.__params).build()
        logger.debug('executing sql [{}] with params \n{}'.format(sql, ','.join([str(param) for param in params])))
        cursor.execute(sql, params)
        return cursor.rowcount


class SelectTableSqlOperation(TableOperation):
    __table_name: str = None
    __params: dict = None
    __columns: list = None

    def __init__(self, cache: TableStructureCache, table_name: str, params: dict = None, columns: list = None):
        super().__init__(cache)
        if columns is None:
            columns = []
        self.__table_name = table_name
        self.__params = params
        self.__columns = columns

    def operate(self, cursor):
        sql_builder = SqlBuilder.select_from(self.__table_name)
        if len(self.__columns) > 0:
            sql_builder.select_column(self.__columns)
        params = []
        if self.__params is not None:
            table_structure = self._table_cache.find_structure(self.__table_name, cursor)
            param_columns = table_structure.exists_columns(self.__params)
            sql_builder.where_equals(param_columns, self.__params)
            params = [self.__params[key] for key in param_columns if self.__params[key] is not None]
        sql = sql_builder.build()
        logger.debug('executing sql [{}] with params \n{}'.format(sql, ','.join([str(param) for param in params])))
        return self._operate(cursor, sql, params)

    def _operate(self, cursor, sql, params):
        return QuerySqlOperation(sql, params).operate(cursor)


class SelectOneFromTableOperation(SelectTableSqlOperation):
    def __init__(self, cache: TableStructureCache, table_name: str, params: dict = None, columns: list = None):
        super().__init__(cache, table_name, params, columns)

    def _operate(self, cursor, sql, params):
        return FindOneOperation(sql, params).operate(cursor)


class SelectObjectFromTableOperation(SelectTableSqlOperation):

    def __init__(self, cache: TableStructureCache, table_name: str, column: str, params: dict = None):
        super().__init__(cache, table_name, params, [column])

    def _operate(self, cursor, sql, params):
        return QueryObjOperation(sql, params).operate(cursor)



from abc import ABCMeta, abstractmethod

__all__ = ['SqlBuilder']


class SqlBuilder:
    __metadata__ = ABCMeta
    _query: list = None

    def __init__(self):
        self._query = []

    def where_equals(self, column: (str, list), params: dict):
        self._check_valid_column(column)
        if isinstance(column, list):
            self.where([self._where_item(col, params) for col in column])
        else:
            self.where(self._where_item(column, params))
        return self

    @staticmethod
    def _where_item(column: str, params: dict):
        val = params.get(column)
        if val is None:
            return f'`{column}` is null'
        else:
            return f'`{column}`=%s'

    def where(self, fragment: (str, list)):
        if isinstance(fragment, list):
            self._query += fragment
        else:
            self._query.append(fragment)
        return self

    def _build_where_fragments(self):
        return ' and '.join([self._wrap_(fragm, '(', ')') for fragm in self._query])

    @staticmethod
    def insert_into(table_name: str):
        return InsertSqlBuilder(table_name)

    @staticmethod
    def update(table_name: str):
        return UpdateSqlBuilder(table_name)

    @staticmethod
    def select_from(table_name: str):
        return SelectSqlBuilder(table_name)

    @staticmethod
    def delete_from(table_name: str):
        return DeleteSqlBuilder(table_name)

    @abstractmethod
    def build(self):
        pass

    @staticmethod
    def _wrap_(content: str, open_symbol: str, close_symbol: str = None):
        if close_symbol is None:
            close_symbol = open_symbol
        return open_symbol + content + close_symbol

    @staticmethod
    def _check_valid_column(column: (str, list)):
        invalid_characters = ['`', '*']
        if isinstance(column, list):
            for col in column:
                if not col.isprintable() or max([col.count(c) for c in invalid_characters]) > 0:
                    raise Exception('invalid column name:{}'.format(col))
        else:
            if column.count('`') > 0:
                raise Exception('invalid column name:{}'.format(column))


class InsertSqlBuilder(SqlBuilder):
    __table = None
    __columns = None

    def __init__(self, table_name: str):
        super().__init__()
        if table_name.count('`') > 0:
            raise Exception('Invalid table name:{}'.format(table_name))
        self.__table = table_name
        self.__columns = []

    def where(self, fragment):
        raise Exception('Unsupported operation')

    def values(self, column: (str, list)):
        self._check_valid_column(column)
        if isinstance(column, list):
            self.__columns += column
        else:
            self.__columns.append(column)
        return self

    def build(self):
        tpl = "insert into `{}` ({}) values ({})"
        return tpl.format(self.__table, ','.join([self._wrap_(col, '`') for col in self.__columns]),
                          ','.join(['%s'] * len(self.__columns)))


class UpdateSqlBuilder(SqlBuilder):
    __table: str = None
    __setters: list = None

    def __init__(self, table: str):
        super().__init__()
        if table.count('`') > 0:
            raise Exception('Invalid table name:{}'.format(table))
        self.__table = table
        self.__setters = []

    def set(self, column: (str, list)):
        self._check_valid_column(column)
        if isinstance(column, list):
            self.__setters += column
        else:
            self.__setters.append(column)
        return self

    def build(self):
        tpl = 'update `{}` set {} where {}'
        return tpl.format(self.__table,
                          ','.join([self.__set_fragment(col) for col in self.__setters]),
                          self._build_where_fragments())

    def __set_fragment(self, column: str):
        return self._wrap_(column, '`') + '=%s'


class SelectSqlBuilder(SqlBuilder):
    __table: str = None
    __selects: list = None

    def __init__(self, table_name: str):
        super().__init__()
        self.__table = table_name
        self.__selects = []

    def select_column(self, column: (str, list)):
        self._check_valid_column(column)
        if isinstance(column, list):
            self.__selects += column
        else:
            self.__selects.append(column)
        return self

    def build(self):
        column_params = ','.join([self._wrap_(col, '`') for col in self.__selects]) if len(self.__selects) > 0 else '*'
        tpl = f'select {column_params} from `{self.__table}` '
        if len(self._query) > 0:
            tpl += 'where ' + self._build_where_fragments()
        return tpl


class DeleteSqlBuilder(SqlBuilder):
    __table: str = None

    def __init__(self, table_name: str):
        super().__init__()
        self.__table = table_name

    def build(self):
        tpl = 'delete from `{}` where {}'
        return tpl.format(self.__table,
                          self._build_where_fragments())



from setuptools import setup, find_packages

setup(
    name='yixsoft-ezmysql',
    version='0.0.2',
    author='yixsoft',
    author_email='davepotter@163.com',
    url='https://gitee.com/yixsoft/python-ezymysql',
    packages=find_packages(),
    description=u'Auto scan table structure and generate sql & execute by dict',
    install_requires=['mysql-connector']
)



def page_to_offset(page: int, limit: int) -> int:
    """
    calculate limit for mysql by page and limit
    :param page: page number, 1-based
    :param limit: page size
    :return: sql limit param
    """
    return max([0, page - 1]) * limit


def total_pages(limit: int, total_count: int) -> int:
    """
    calculate total pages by limit and total count
    :param limit: page size
    :param total_count: total record count
    :return: total page count
    """
    return int((total_count - 1) / limit) + 1



import logging
from abc import ABCMeta, abstractmethod

logger = logging.getLogger('ezmysql')

__all__ = ['SqlOperation', 'UpdateSqlOperation', 'DeleteSqlOperation', 'CommonInsertOperation', 'QuerySqlOperation',
           'QueryObjOperation', 'DescTableOperation', 'FindOneOperation']


class SqlOperation:
    __metadata__ = ABCMeta

    @abstractmethod
    def operate(self, cursor):
        pass

    @staticmethod
    def _convert_value(value):
        if isinstance(value, bytearray):
            return value.decode('utf-8')
        return value


class UpdateSqlOperation(SqlOperation):
    __sql = None
    __params = None

    def __init__(self, sql: str, params: list = None):
        if params is None:
            params = []
        self.__sql = sql
        self.__params = params

    def operate(self, cursor):
        logger.debug(f'executing sql----->{self.__sql}')
        if self.__params:
            logger.debug(f'params------->{",".join([str(v) for v in self.__params])}')
        cursor.execute(self.__sql, self.__params)
        rc = cursor.rowcount
        logger.debug(f'updated row count<--------{rc}')
        return rc


class DeleteSqlOperation(UpdateSqlOperation):

    def __init__(self, sql: str, params: list = None):
        super().__init__(sql, params)


class CommonInsertOperation(UpdateSqlOperation):
    def __init__(self, sql: str, params: dict = None):
        super().__init__(sql, params)

    def operate(self, cursor):
        super().operate(cursor)
        return cursor.lastrowid


class QuerySqlOperation(SqlOperation):
    __sql = None
    __params = None
    __lowercase = False

    def __init__(self, sql: str, params: list = None, lowercase=False):
        if params is None:
            params = []
        sql = sql.strip()
        if not sql.upper().startswith('SELECT') and not sql.upper().startswith('DESC'):
            raise Exception('Invalid call query:{}'.format(sql))
        self.__sql = sql
        self.__params = params
        self.__lowercase = lowercase

    def operate(self, cursor):
        logger.debug(f'executing sql----->{self.__sql}')
        if self.__params:
            logger.debug(f'params------->{",".join([str(v) for v in self.__params])}')
        cursor.execute(self.__sql, self.__params)
        try:
            fetchall = cursor.fetchall()
            logger.debug(f'response rows<---------{len(fetchall)}')
            if not self.__lowercase:
                return fetchall
            else:
                return [self._lower_key(item) for item in fetchall]
        except any:
            return []

    @staticmethod
    def _lower_key(obj):
        return dict([(k.lower(), obj[k]) for k in obj])


class FindOneOperation(QuerySqlOperation):

    def __init__(self, sql: str, params: list = None, lowercase=False):
        super().__init__(sql, params, lowercase)

    def operate(self, cursor):
        data = super().operate(cursor)
        if len(data) > 1:
            raise Exception('find one get more than 1 result', self.__sql, self.__params)
        return None if len(data) == 0 else data[0]


class QueryObjOperation(FindOneOperation):
    def __init__(self, sql: str, params: list):
        super().__init__(sql, params)

    def operate(self, cursor):
        data = super().operate(cursor)
        if data is None:
            return None
        items = data.items()
        if len(items) > 1:
            raise Exception('Query object but result contains more than one key:{}'.format(data))
        return next(iter(data.values()))


class DescTableOperation(QuerySqlOperation):
    def __init__(self, table_name: str):
        if table_name.count('`') > 0:
            raise Exception('invalid table name:{}'.format(table_name))
        super().__init__('DESC `{}`'.format(table_name), lowercase=True)



import logging
import os
import threading
from datetime import timedelta, datetime
from typing import List, Optional

from mysql.connector import errors
from mysql.connector.pooling import CNX_POOL_MAXSIZE
from mysql.connector.pooling import MySQLConnectionPool, PooledMySQLConnection

from .operation import *
from .TableCache import TableStructureCache

CONNECTION_POOL_LOCK = threading.RLock()
logger = logging.getLogger('ezmysql')


class Pool(MySQLConnectionPool):
    __table_cache: TableStructureCache = None

    def __init__(self, cache_timeout: timedelta = timedelta(minutes=10), **kwargs):
        kwargs['pool_reset_session'] = False
        super().__init__(**kwargs)
        self.__table_cache = TableStructureCache(cache_timeout)

    @staticmethod
    def from_os_env():
        os_zone = datetime.now().astimezone().strftime('%z')
        mysql_zone = os_zone[0:-2] + ':' + os_zone[-2:]
        conn_params = dict(
            host=os.environ.get('APP_DB_HOST') or '127.0.0.1',
            port=os.environ.get('APP_DB_PORT') or 3306,
            database=os.environ.get('APP_DB_SCHEMA'),
            user=os.environ.get('APP_DB_USER'),
            password=os.environ.get('APP_DB_PWD'),
            charset='utf8mb4',
            autocommit=True,
            buffered=True,
            time_zone=os.environ.get('APP_DB_TIMEZONE') or mysql_zone,
            pool_reset_session=False
        )
        conn_params = {k: v for k, v in conn_params.items() if v is not None}
        return Pool(pool_reset_session=False, pool_name='{}_{}'.format(conn_params['database'], conn_params['user']),
                    **conn_params)

    def connect(self):
        try:
            return self.get_connection()
        except errors.PoolError:
            # Pool size should be lower or equal to CNX_POOL_MAXSIZE
            if self.pool_size < CNX_POOL_MAXSIZE:
                with threading.Lock():
                    new_pool_size = self.pool_size + 1
                    try:
                        self._set_pool_size(new_pool_size)
                        self._cnx_queue.maxsize = new_pool_size
                        self.add_connection()
                    except Exception as e:
                        logging.exception(e)
                    return self.connect()
            else:
                with CONNECTION_POOL_LOCK:
                    cnx = self._cnx_queue.get(block=True)
                    if not cnx.is_connected() or self._config_version != cnx._pool_config_version:
                        cnx.config(**self._cnx_config)
                        try:
                            cnx.reconnect()
                        except errors.InterfaceError:
                            # Failed to reconnect, give connection back to pool
                            self._queue_connection(cnx)
                            raise
                        cnx._pool_config_version = self._config_version
                    return PooledMySQLConnection(self, cnx)
        except Exception:
            raise

    def execute_with_cursor(self, operation: SqlOperation):
        cnx = cursor = None
        try:
            cnx = self.connect()
            cursor = cnx.cursor(buffered=True, dictionary=True)
            return operation.operate(cursor)
        except Exception:
            raise
        finally:
            if cursor:
                cursor.close()
            if cnx:
                cnx.close()

    def query(self, sql: str, params: list = None) -> List[dict]:
        """
        common select, return a list with dict
        :param sql: select sql
        :param params: args list
        :return: dict objects in array
        """
        query_operation = QuerySqlOperation(sql, params)
        return self.execute_with_cursor(query_operation)

    def query_from_table(self, table: str, columns: list = None, param: dict = None) -> List[dict]:
        """
        execute simple select on a table: select <columns> from <table> where (<key1>=<value1>) and (<key2>=<value2>)
        :param table: table name
        :param columns: select columns. None means select * from table.
        :param param: args in where part. column=value mode. will ignore column if not exists in table
        :return: matching results in list. result item is a dict
        """
        op = SelectTableSqlOperation(self.__table_cache, table, param, columns)
        return self.execute_with_cursor(op)

    def findone(self, sql: str, params: list = None) -> Optional[dict]:
        """
        if a select sql has one row max. can use this method. which will response the single row or None if not exists
        :param sql: select sql
        :param params: where args
        :return: single row in dict or None if not found. if exists multiple results will raise an exception
        """
        query_operation = FindOneOperation(sql, params)
        return self.execute_with_cursor(query_operation)

    def findone_from_table(self, table: str, columns: list = None, param: dict = None) -> dict:
        """
        find one in table mode
        :param table: table name
        :param columns: select columns. None means select *
        :param param: args in where part. column=value mode. will ignore column if not exists in table
        :return: single row in dict or None if not found. if exists multiple results will raise an exception
        """
        op = SelectOneFromTableOperation(self.__table_cache, table, param, columns)
        return self.execute_with_cursor(op)

    def query_obj(self, sql: str, params: list = None) -> Optional:
        """
        similar to findone, if select has only one column, can use this method to get value directly
        :param sql: select sql, must have only 1 column and max 1 row to response
        :param params: args list
        :return: column value or None if not found
        """
        query_operation = QueryObjOperation(sql, params)
        return self.execute_with_cursor(query_operation)

    def query_obj_from_table(self, table: str, column: str, param: dict = None) -> Optional:
        """
        select only one column from table.
        :param table: table name
        :param column: required column
        :param param: args in where part. column=value mode. will ignore column if not exists in table
        :return: value of determined column or None if not found
        """
        op = SelectObjectFromTableOperation(self.__table_cache, table, column, param)
        return self.execute_with_cursor(op)

    def update(self, sql: str, params: list = None) -> int:
        """
        process update sql
        :param sql: update sql
        :param params: args list
        :return: affected row count
        """
        query_operation = UpdateSqlOperation(sql, params)
        return self.execute_with_cursor(query_operation)

    def delete(self, sql: str, params: list = None) -> int:
        """
        process delete sql
        :param sql: delete sql
        :param params: args list
        :return: affected row count
        """
        del_operation = DeleteSqlOperation(sql, params)
        return self.execute_with_cursor(del_operation)

    def update_table(self, table_name: str, updates: dict, where: dict) -> int:
        """
        auto generate update sql for determined table and execute
        :param table_name: table name
        :param updates: args in set part. key=value, will ignore key if not exists in table columns
        :param where: args in where part. column=value mode. will ignore column if not exists in table
        :return: affected row count
        """
        op = UpdateTableOperation(self.__table_cache, table_name, updates, where)
        return self.execute_with_cursor(op)

    def delete_table(self, table_name: str, where: dict) -> int:
        """
        auto generate delete sql for determined table and execute
        :param table_name: table name
        :param where: args in where part. column=value mode. will ignore column if not exists in table
        :return: affected row count
        """
        op = DeleteTableOperation(self.__table_cache, table_name, where)
        return self.execute_with_cursor(op)

    def insert_table(self, table_name: str, item: dict):
        """
        insert the dict into determined table
        :param table_name: table name
        :param item: key=value in table values, will ignore columns if not exists in table
        :return: primary key value
        """
        query_operation = TableInsertOperation(self.__table_cache, table_name, item)
        return self.execute_with_cursor(query_operation)

    def insert(self, sql: str, params: list = None):
        """
        process insert sql
        :param sql: insert sql
        :param params: args list
        :return: primary key value
        """
        query_operation = CommonInsertOperation(sql, params)
        return self.execute_with_cursor(query_operation)

    def begin(self, consistent_snapshot=False, isolation_level=None, readonly=None):
        cnx = self.connect()
        cnx.start_transaction(consistent_snapshot, isolation_level, readonly)
        return Transaction(self.__table_cache, cnx)


class Transaction(object):

    def __init__(self, cache: TableStructureCache, connection):
        self.__table_cache = cache
        self.cnx = None
        if isinstance(connection, PooledMySQLConnection):
            self.cnx = connection
            self.cursor = connection.cursor(buffered=True, dictionary=True)
        else:
            raise AttributeError("connection should be a PooledMySQLConnection")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None and exc_val is None and exc_tb is None:
            self.commit()
        else:
            # will raise with-body's Exception, should deal with it
            self.rollback()
        self.close()

    def query(self, sql: str, params: list = None):
        """
        common select, return a list with dict
        :param sql: select sql
        :param params: args list
        :return: dict objects in array
        """
        query_operation = QuerySqlOperation(sql, params)
        return query_operation.operate(self.cursor)

    def query_from_table(self, table: str, columns: list = None, param: dict = None):
        """
        execute simple select on a table: select <columns> from <table> where (<key1>=<value1>) and (<key2>=<value2>)
        :param table: table name
        :param columns: select columns. None means select * from table.
        :param param: args in where part. column=value mode. will ignore column if not exists in table
        :return: matching results in list. result item is a dict
        """
        op = SelectTableSqlOperation(self.__table_cache, table, param, columns)
        return op.operate(self.cursor)

    def findone(self, sql: str, params: list = None):
        """
        if a select sql has one row max. can use this method. which will response the single row or None if not exists
        :param sql: select sql
        :param params: where args
        :return: single row in dict or None if not found. if exists multiple results will raise an exception
        """
        query_operation = FindOneOperation(sql, params)
        return query_operation.operate(self.cursor)

    def findone_from_table(self, table: str, columns: list = None, param: dict = None):
        """
        find one in table mode
        :param table: table name
        :param columns: select columns. None means select *
        :param param: args in where part. column=value mode. will ignore column if not exists in table
        :return: single row in dict or None if not found. if exists multiple results will raise an exception
        """
        op = SelectOneFromTableOperation(self.__table_cache, table, param, columns)
        return op.operate(self.cursor)

    def query_obj(self, sql: str, params: list = None):
        """
        similar to findone, if select has only one column, can use this method to get value directly
        :param sql: select sql, must have only 1 column and max 1 row to response
        :param params: args list
        :return: column value or None if not found
        """
        query_operation = QueryObjOperation(sql, params)
        return query_operation.operate(self.cursor)

    def query_obj_from_table(self, table: str, column: str, param: dict = None):
        """
        select only one column from table.
        :param table: table name
        :param column: required column
        :param param: args in where part. column=value mode. will ignore column if not exists in table
        :return: value of determined column or None if not found
        """
        op = SelectObjectFromTableOperation(self.__table_cache, table, column, param)
        return op.operate(self.cursor)

    def insert_table(self, table_name: str, item: dict):
        """
        insert the dict into determined table
        :param table_name: table name
        :param item: key=value in table values, will ignore columns if not exists in table
        :return: primary key value
        """
        cursor = self.cursor
        op = TableInsertOperation(self.__table_cache, table_name, item)
        return op.operate(cursor)

    def insert(self, sql: str, params: list = None):
        """
        process insert sql
        :param sql: insert sql
        :param params: args list
        :return: primary key value
        """
        cursor = self.cursor
        op = CommonInsertOperation(sql, params)
        return op.operate(cursor)

    def update(self, sql: str, params: list = None):
        """
        process update sql
        :param sql: update sql
        :param params: args list
        :return: affected row count
        """
        cursor = self.cursor
        op = UpdateSqlOperation(sql, params)
        return op.operate(cursor)

    def delete(self, sql: str, params: list = None):
        """
        process delete sql
        :param sql: delete sql
        :param params: args list
        :return: affected row count
        """
        cursor = self.cursor
        op = DeleteSqlOperation(sql, params)
        return op.operate(cursor)

    def update_table(self, table_name: str, updates: dict, where: dict):
        """
        auto generate update sql for determined table and execute
        :param table_name: table name
        :param updates: args in set part. key=value, will ignore key if not exists in table columns
        :param where: args in where part. column=value mode. will ignore column if not exists in table
        :return: affected row count
        """
        cursor = self.cursor
        op = UpdateTableOperation(self.__table_cache, table_name, updates, where)
        return op.operate(cursor)

    def delete_table(self, table_name: str, where: dict):
        """
        auto generate delete sql for determined table and execute
        :param table_name: table name
        :param where: args in where part. column=value mode. will ignore column if not exists in table
        :return: affected row count
        """
        cursor = self.cursor
        op = DeleteTableOperation(self.__table_cache, table_name, where)
        return op.operate(cursor)

    def commit(self):
        self.cnx.commit()

    def rollback(self):
        self.cnx.rollback()

    def close(self):
        self.cursor.close()
        self.cnx.close()
